# This is the base class for the generic decorator
class Decorator:
    def __init__(self, *args, **kwargs):
        from functools import update_wrapper, WRAPPER_ASSIGNMENTS, WRAPPER_UPDATES
        self._function = None
        self._config = None
        if (not kwargs) and (len(args) == 1) and callable(args[0]):
            self._function = args[0]
            update_wrapper(
                self,
                self._function,
                assigned=WRAPPER_ASSIGNMENTS,
                updated=WRAPPER_UPDATES)
        else:
            self._config = (args, kwargs)

    def __call__(self, *args, **kwargs):
        if self._function:
            return self.call(self._config, self._function, *args, **kwargs)
        else:
            obj = self.__class__(*args, **kwargs)
            obj._config = self._config
            return obj

    def call(self, config, function, *args, **kwargs):
        return function(*args, **kwargs)


# this is out implementation
class decorator(Decorator):
    def call(self, config, function, *args, **kwargs):
        if config is None:
            factor = 2              # form-1 config is None
        elif not config[0]:
            factor = 3              # form-2 config is ((), {}) 2-tuple
        else:
            factor = config[1]["b"] # form-3 config is ((1, 2), { "a" : 3, "b": 4 }) 2-tuple

        return function(*args, **kwargs) * factor


# this is the "test"

def function(x, y):
    return x+y**2


@decorator
def fn1(x, y):
    "form-1 decorator"
    return function(x, y)


@decorator()
def fn2(x, y):
    "form-2 decorator"
    return function(x, y)


@decorator(1, 2, a=3, b=4)
def fn3(x, y):
    "form-3 decorator"
    return function(x, y)


print(fn1.__doc__, fn1(1, 2))
print(fn2.__doc__, fn2(1, 2))
print(fn3.__doc__, fn3(1, 2))
